#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define ARRAY_SIZE(a)	(sizeof(a)/sizeof((a)[0]))

#define ALPH_SPECIAL	0x0001
#define ALPH_LOWER	0x0002
#define ALPH_UPPER	0x0004
#define ALPH_DIGITS	0x0008
#define ALPH_ALNUM	(ALPH_LOWER | ALPH_UPPER | ALPH_DIGITS)
#define ALPH_ALL	(ALPH_SPECIAL | ALPH_ALNUM)

/* 
 * Generates ASCII character set selected by flags.
 * 
 * result: returned result, user is responsible to free(3) it
 * size:   returned size (in characters) of result string
 * flags:  flags selecting character subset
 *
 * Returns: 0 on success, -1 on failure.
 */
int
gen_alphabet(char **result, int *size, int flags)
{
	char *res, *tmp;
	char *special = "!@#$%^&*()-=_+|\\/?.,\";:\'~<>{}[]"; 
	char *lower = "abcdefghijklmnopqrstuvwxyz";
	char *upper = "ABCDEFGHIJKLMNOPQRSTRUVWXYZ";
	char *digits = "1234567890";
	int i, len, n;

	struct {
		char	*set;
		int	flag;
	} *ctp, ctab[] = {
		{ special, ALPH_SPECIAL },
		{ lower,   ALPH_LOWER },
		{ upper,   ALPH_UPPER },
		{ digits,  ALPH_DIGITS },
	};

	len = 0;
	res = NULL;

	for (i = 0; i < ARRAY_SIZE(ctab); i++) {

		ctp = ctab + i;

		if ((ctp->flag & flags) == 0)
			continue;
		
		n = strlen(ctp->set) + len;
		tmp = realloc(res, n + 1);
		if (tmp == NULL) {
			free(res);
			return -1;
		}
		res = tmp;
		memcpy(res + len, ctp->set, strlen(ctp->set));
		len = n;
	}

	if (res != NULL && len > 0)
		res[len] = '\0';

	if (size != NULL)
		*size = len;

	if (result != NULL)
		*result = res;
	else
		free(res);	

	return 0;
}

int
main()
{
	char *str;
	int res, n, i;

	int flags[] = {
		ALPH_SPECIAL | ALPH_DIGITS,
		ALPH_SPECIAL | ALPH_LOWER | ALPH_UPPER,
		ALPH_ALL
	};
	char *names[] = { 
		"ALPH_SPECIAL | ALPH_DIGITS",
		"ALPH_SPECIAL | ALPH_LOWER | ALPH_UPPER",
		"ALPH_ALL",
		NULL
	};

	for (i = 0; i < ARRAY_SIZE(flags); i++) {

		printf("Generating alphabet for: %s\n", names[i]);

		res = gen_alphabet(&str, &n, flags[i]);

		if (res == -1)
			exit(1);

		printf("  %d chars: %s\n", n, str);
		free(str);
	}

	return 0;
}

